@@ -1,9 +1,12 @@
 from typing import Iterable
+from typing import Tuple
 from typing import Union
-
+import logging
 import numpy as np
 import torch
+import logging
 
+import mp_pytorch.util
 from mp_pytorch.basis_gn import ProDMPBasisGenerator
 from .promp import ProMP
 
@@ -37,6 +40,13 @@ class ProDMP(ProMP):
             raise ValueError(
                 f'ProDMP requires a ProDMP basis generator, {type(basis_gn)} is not supported.')
 
+        # Disable learning of weights or goal
+        self.disable_weights = kwargs.get("disable_weights", False)
+        self.disable_goal = kwargs.get("disable_goal", False)
+        assert not (self.disable_weights and self.disable_goal), \
+            "Cannot disable both weights and goal learning."
+
+        # Super init
         super().__init__(basis_gn, num_dof, weights_scale, dtype, device,
                          **kwargs)
 
@@ -46,34 +56,56 @@ class ProDMP(ProMP):
         # Goal scale
         self.auto_scale_basis = kwargs.get("auto_scale_basis", False)
         self.goal_scale = goal_scale
-        self.weights_goal_scale = self.get_weights_goal_scale(self.auto_scale_basis)
+        self.weights_goal_scale = self.get_weights_goal_scale(
+            self.auto_scale_basis)
 
         # Runtime intermediate variables shared by different getting functions
         self.y1 = None
         self.y2 = None
         self.dy1 = None
         self.dy2 = None
-        self.y1_bc = None
-        self.y2_bc = None
-        self.dy1_bc = None
-        self.dy2_bc = None
+        self.y1_init = None
+        self.y2_init = None
+        self.dy1_init = None
+        self.dy2_init = None
 
-        self.pos_bc = None
-        self.vel_bc = None
+        self.pos_init = None
+        self.vel_init = None
         self.pos_H_single = None
         self.vel_H_single = None
 
         self.pos_H_multi = None
         self.vel_H_multi = None
 
+        # Padding params when weights or goal is disabled
+        padding_basis = self.basis_gn.num_basis if self.disable_weights else 0
+        padding_goal = 1 if self.disable_goal else 0
+        if self.disable_weights or self.disable_goal:
+            self.padding = torch.nn.ConstantPad2d(
+                (padding_basis, padding_goal, 0, 0), 0)
+            logging.warning(
+                "Padding ProDMP is being used. Only the traj position"
+                " and velocity can be computed correctly. The other "
+                "entities are not guaranteed.")
+        else:
+            self.padding = lambda x: x
+
     @property
     def _num_local_params(self) -> int:
         """
         Returns: number of parameters of current class
         """
-        return super()._num_local_params + self.num_dof
+        if self.disable_weights:
+            # Use goal only
+            return self.num_dof
+        if self.disable_goal:
+            # Use basis only
+            return super()._num_local_params
+        else:
+            # Use both basis and goal
+            return super()._num_local_params + self.num_dof
 
-    def get_weights_goal_scale(self, auto_scale_basis = False) -> torch.Tensor:
+    def get_weights_goal_scale(self, auto_scale_basis=False) -> torch.Tensor:
         """
         Compute scaling factors of weights and goal
         Args:
@@ -113,48 +145,54 @@ class ProDMP(ProMP):
 
         super().set_times(times)
 
-    def set_boundary_conditions(self, bc_time: Union[torch.Tensor, np.ndarray],
-                                bc_pos: Union[torch.Tensor, np.ndarray],
-                                bc_vel: Union[torch.Tensor, np.ndarray]):
+    def set_initial_conditions(self, init_time: Union[torch.Tensor, np.ndarray],
+                               init_pos: Union[torch.Tensor, np.ndarray],
+                               init_vel: Union[torch.Tensor, np.ndarray]):
         """
-        Set boundary conditions in a batched manner
+        Set initial conditions in a batched manner
 
         Args:
-            bc_time: boundary condition time
-            bc_pos: boundary condition position
-            bc_vel: boundary condition velocity
+            init_time: initial condition time
+            init_pos: initial condition position
+            init_vel: initial condition velocity
 
         Returns:
             None
         """
-        # Shape of bc_time:
+        # Shape of init_time:
         # [*add_dim]
         #
-        # Shape of bc_pos:
+        # Shape of init_pos:
         # [*add_dim, num_dof]
         #
-        # Shape of bc_vel:
+        # Shape of init_vel:
         # [*add_dim, num_dof]
 
-        bc_time = torch.as_tensor(bc_time, dtype=self.dtype, device=self.device)
-        bc_pos = torch.as_tensor(bc_pos, dtype=self.dtype, device=self.device)
-        bc_vel = torch.as_tensor(bc_vel, dtype=self.dtype, device=self.device)
+        init_time = torch.as_tensor(init_time, dtype=self.dtype,
+                                    device=self.device)
+        init_pos = torch.as_tensor(init_pos, dtype=self.dtype,
+                                   device=self.device)
+        init_vel = torch.as_tensor(init_vel, dtype=self.dtype,
+                                   device=self.device)
 
-        assert list(bc_time.shape) == [*self.add_dim]
-        assert list(bc_pos.shape) == list(bc_vel.shape) and list(bc_vel.shape) == [*self.add_dim, self.num_dof]
+        assert list(init_time.shape) == [*self.add_dim]
+        assert list(init_pos.shape) == list(init_vel.shape) and list(
+            init_vel.shape) == [*self.add_dim, self.num_dof]
 
-        bc_time = torch.as_tensor(bc_time, dtype=self.dtype, device=self.device)
-        y1_bc, y2_bc, dy1_bc, dy2_bc = self.basis_gn.general_solution_values(bc_time[..., None])
+        init_time = torch.as_tensor(init_time, dtype=self.dtype,
+                                    device=self.device)
+        y1_init, y2_init, dy1_init, dy2_init = self.basis_gn.general_solution_values(
+            init_time[..., None])
 
-        self.y1_bc = y1_bc.squeeze(-1)
-        self.y2_bc = y2_bc.squeeze(-1)
-        self.dy1_bc = dy1_bc.squeeze(-1)
-        self.dy2_bc = dy2_bc.squeeze(-1)
+        self.y1_init = y1_init.squeeze(-1)
+        self.y2_init = y2_init.squeeze(-1)
+        self.dy1_init = dy1_init.squeeze(-1)
+        self.dy2_init = dy2_init.squeeze(-1)
 
-        super().set_boundary_conditions(bc_time, bc_pos, bc_vel)
+        super().set_initial_conditions(init_time, init_pos, init_vel)
 
     def get_traj_pos(self, times=None, params=None,
-                     bc_time=None, bc_pos=None, bc_vel=None,
+                     init_time=None, init_pos=None, init_vel=None,
                      flat_shape=False):
         """
         Compute trajectory pos
@@ -164,9 +202,9 @@ class ProDMP(ProMP):
         Args:
             times: time points
             params: learnable parameters
-            bc_time: boundary condition time
-            bc_pos: boundary condition position
-            bc_vel: boundary condition velocity
+            init_time: initial condition time
+            init_pos: initial condition position
+            init_vel: initial condition velocity
             flat_shape: if flatten the dimensions of Dof and time
 
         Returns:
@@ -177,7 +215,7 @@ class ProDMP(ProMP):
         # [*add_dim, num_times, num_dof] or [*add_dim, num_dof * num_times]
 
         # Update inputs
-        self.update_inputs(times, params, None, bc_time, bc_pos, bc_vel)
+        self.update_inputs(times, params, None, init_time, init_pos, init_vel)
 
         # Reuse result if existing
         if self.pos is not None:
@@ -188,10 +226,9 @@ class ProDMP(ProMP):
             # Recompute otherwise
             self.compute_intermediate_terms_single()
 
-            # Reshape params from [*add_dim, num_dof * num_basis_g]
-            # to [*add_dim, num_dof, num_basis_g]
-            params = self.params.reshape(
-                [*self.add_dim, self.num_dof, -1])
+            # Reshape (and pad) params to [*add_dim, num_dof, num_basis_g]
+            params = self.params.reshape([*self.add_dim, self.num_dof, -1])
+            params = self.padding(params)
 
             # Scale basis functions
             pos_H_single = self.pos_H_single * self.weights_goal_scale
@@ -204,7 +241,7 @@ class ProDMP(ProMP):
             pos_linear = \
                 torch.einsum('...jk,...ik->...ij', pos_H_single, params)
             pos_linear = torch.reshape(pos_linear, [*self.add_dim, -1])
-            pos = self.pos_bc + pos_linear
+            pos = self.pos_init + pos_linear
             self.pos = pos
 
         if not flat_shape:
@@ -216,8 +253,8 @@ class ProDMP(ProMP):
 
         return pos
 
-    def get_traj_pos_cov(self, times=None, params_L=None, bc_time=None,
-                         bc_pos=None, bc_vel=None, reg: float = 1e-4):
+    def get_traj_pos_cov(self, times=None, params_L=None, init_time=None,
+                         init_pos=None, init_vel=None, reg: float = 1e-4):
         """
         Compute and return position covariance
 
@@ -226,9 +263,9 @@ class ProDMP(ProMP):
         Args:
             times: time points
             params_L: learnable parameters' variance
-            bc_time: boundary condition time
-            bc_pos: boundary condition position
-            bc_vel: boundary condition velocity
+            init_time: initial condition time
+            init_pos: initial condition position
+            init_vel: initial condition velocity
             reg: regularization term
 
         Returns:
@@ -239,7 +276,7 @@ class ProDMP(ProMP):
         # [*add_dim, num_dof * num_times, num_dof * num_times]
 
         # Update inputs
-        self.update_inputs(times, None, params_L, bc_time, bc_pos, bc_vel)
+        self.update_inputs(times, None, params_L, init_time, init_pos, init_vel)
 
         # Reuse result if existing
         if self.pos_cov is not None:
@@ -270,13 +307,15 @@ class ProDMP(ProMP):
                                               pos_cov)).item() * traj_cov_reg
 
         # Add regularization term for numerical stability
-        self.pos_cov = pos_cov + torch.eye(pos_cov.shape[-1]) * reg_term_pos
+        self.pos_cov = pos_cov + torch.eye(pos_cov.shape[-1],
+                                           dtype=self.dtype,
+                                           device=self.device) * reg_term_pos
 
         return self.pos_cov
 
-    def get_traj_pos_std(self, times=None, params_L=None, bc_time=None,
-                         bc_pos=None,
-                         bc_vel=None, flat_shape=False, reg: float = 1e-4):
+    def get_traj_pos_std(self, times=None, params_L=None, init_time=None,
+                         init_pos=None,
+                         init_vel=None, flat_shape=False, reg: float = 1e-4):
         """
         Compute trajectory standard deviation
 
@@ -285,9 +324,9 @@ class ProDMP(ProMP):
         Args:
             times: time points
             params_L: learnable parameters' variance
-            bc_time: boundary condition time
-            bc_pos: boundary condition position
-            bc_vel: boundary condition velocity
+            init_time: initial condition time
+            init_pos: initial condition position
+            init_vel: initial condition velocity
             flat_shape: if flatten the dimensions of Dof and time
             reg: regularization term
 
@@ -299,7 +338,7 @@ class ProDMP(ProMP):
         # [*add_dim, num_times, num_dof] or [*add_dim, num_dof * num_times]
 
         # Update inputs
-        self.update_inputs(times, None, params_L, bc_time, bc_pos, bc_vel)
+        self.update_inputs(times, None, params_L, init_time, init_pos, init_vel)
 
         # Reuse result if existing
         if self.pos_std is not None:
@@ -326,7 +365,7 @@ class ProDMP(ProMP):
         return pos_std
 
     def get_traj_vel(self, times=None, params=None,
-                     bc_time=None, bc_pos=None, bc_vel=None,
+                     init_time=None, init_pos=None, init_vel=None,
                      flat_shape=False):
         """
         Compute trajectory velocity
@@ -336,9 +375,9 @@ class ProDMP(ProMP):
         Args:
             times: time points
             params: learnable parameters
-            bc_time: boundary condition time
-            bc_pos: boundary condition position
-            bc_vel: boundary condition velocity
+            init_time: initial condition time
+            init_pos: initial condition position
+            init_vel: initial condition velocity
             flat_shape: if flatten the dimensions of Dof and time
 
         Returns:
@@ -349,7 +388,7 @@ class ProDMP(ProMP):
         # [*add_dim, num_times, num_dof] or [*add_dim, num_dof * num_times]
 
         # Update inputs
-        self.update_inputs(times, params, None, bc_time, bc_pos, bc_vel)
+        self.update_inputs(times, params, None, init_time, init_pos, init_vel)
 
         # Reuse result if existing
         if self.vel is not None:
@@ -358,9 +397,9 @@ class ProDMP(ProMP):
             # Recompute otherwise
             self.compute_intermediate_terms_single()
 
-            # Reshape params from [*add_dim, num_dof * num_basis_g]
-            # to [*add_dim, num_dof, num_basis_g]
+            # Reshape (and pad) params to [*add_dim, num_dof, num_basis_g]
             params = self.params.reshape([*self.add_dim, self.num_dof, -1])
+            params = self.padding(params)
 
             # Scale basis functions
             vel_H_single = self.vel_H_single * self.weights_goal_scale
@@ -373,7 +412,7 @@ class ProDMP(ProMP):
             vel_linear = \
                 torch.einsum('...jk,...ik->...ij', vel_H_single, params)
             vel_linear = torch.reshape(vel_linear, [*self.add_dim, -1])
-            vel = self.vel_bc + vel_linear
+            vel = self.vel_init + vel_linear
 
             # Unscale velocity to original time scale space
             vel = vel / self.phase_gn.tau[..., None]
@@ -388,9 +427,9 @@ class ProDMP(ProMP):
 
         return vel
 
-    def get_traj_vel_cov(self, times=None, params_L=None, bc_time=None,
-                         bc_pos=None,
-                         bc_vel=None, reg: float = 1e-4):
+    def get_traj_vel_cov(self, times=None, params_L=None, init_time=None,
+                         init_pos=None,
+                         init_vel=None, reg: float = 1e-4):
         """
         Get trajectory velocity covariance
 
@@ -399,9 +438,9 @@ class ProDMP(ProMP):
         Args:
             times: time points
             params_L: learnable parameters' variance
-            bc_time: boundary condition time
-            bc_pos: boundary condition position
-            bc_vel: boundary condition velocity
+            init_time: initial condition time
+            init_pos: initial condition position
+            init_vel: initial condition velocity
             reg: regularization term
 
         Returns:
@@ -412,7 +451,7 @@ class ProDMP(ProMP):
         # [*add_dim, num_dof * num_times, num_dof * num_times]
 
         # Update inputs
-        self.update_inputs(times, None, params_L, bc_time, bc_pos, bc_vel)
+        self.update_inputs(times, None, params_L, init_time, init_pos, init_vel)
 
         # Reuse result if existing
         if self.vel_cov is not None:
@@ -444,16 +483,18 @@ class ProDMP(ProMP):
                                               vel_cov)).item() * traj_cov_reg
 
         # Add regularization term for numerical stability
-        vel_cov = vel_cov + torch.eye(vel_cov.shape[-1]) * reg_term_vel
+        vel_cov = vel_cov + torch.eye(vel_cov.shape[-1],
+                                      dtype=self.dtype,
+                                      device=self.device) * reg_term_vel
 
         # Unscale velocity to original time scale space
         self.vel_cov = vel_cov / self.phase_gn.tau[..., None, None] ** 2
 
         return self.vel_cov
 
-    def get_traj_vel_std(self, times=None, params_L=None, bc_time=None,
-                         bc_pos=None,
-                         bc_vel=None, flat_shape=False, reg: float = 1e-4):
+    def get_traj_vel_std(self, times=None, params_L=None, init_time=None,
+                         init_pos=None,
+                         init_vel=None, flat_shape=False, reg: float = 1e-4):
         """
         Compute trajectory velocity standard deviation
 
@@ -462,9 +503,9 @@ class ProDMP(ProMP):
         Args:
             times: time points
             params_L: learnable parameters' variance
-            bc_time: boundary condition time
-            bc_pos: boundary condition position
-            bc_vel: boundary condition velocity
+            init_time: initial condition time
+            init_pos: initial condition position
+            init_vel: initial condition velocity
             flat_shape: if flatten the dimensions of Dof and time
             reg: regularization term
 
@@ -477,7 +518,7 @@ class ProDMP(ProMP):
         # [*add_dim, num_times, num_dof] or [*add_dim, num_dof * num_times]
 
         # Update inputs
-        self.update_inputs(times, None, params_L, bc_time, bc_pos, bc_vel)
+        self.update_inputs(times, None, params_L, init_time, init_pos, init_vel)
 
         # Reuse result if existing
         if self.vel_std is not None:
@@ -503,22 +544,23 @@ class ProDMP(ProMP):
 
     def learn_mp_params_from_trajs(self, times: torch.Tensor,
                                    trajs: torch.Tensor,
-                                   reg: float = 1e-9) -> dict:
+                                   reg: float = 1e-9, **kwargs) -> dict:
         """
         Learn DMP weights and goal given trajectory position
-        Use the initial position and velocity as boundary condition
+        Use the initial position and velocity as initial condition
 
         Args:
             times: trajectory time points
             trajs: trajectory position in batch
             reg: regularization term
+            kwargs: keyword arguments
 
         Returns:
             param_dict: dictionary of parameters containing
                 - params (w + g)
-                - bc_time
-                - bc_pos
-                - bc_vel
+                - init_time
+                - init_pos
+                - init_vel
         """
         # Shape of times:
         # [*add_dim, num_times]
@@ -535,16 +577,24 @@ class ProDMP(ProMP):
 
         trajs = torch.as_tensor(trajs, dtype=self.dtype, device=self.device)
 
-        # Get boundary conditions
+        # Get initial conditions
         dt = self.basis_gn.scaled_dt * self.phase_gn.tau
-        bc_time = times[..., 0]
-        bc_pos = trajs[..., 0, :]
-        bc_vel = torch.diff(trajs, dim=-2)[..., 0, :] / dt
+
+        if all([key in kwargs.keys()
+                for key in ["init_time", "init_pos", "init_vel"]]):
+            logging.warning("ProDMP uses the given initial conditions")
+            init_time = kwargs["init_time"]
+            init_pos = kwargs["init_pos"]
+            init_vel = kwargs["init_vel"]
+        else:
+            init_time = times[..., 0]
+            init_pos = trajs[..., 0, :]
+            init_vel = torch.diff(trajs, dim=-2)[..., 0, :] / dt
 
         # Setup stuff
         self.set_add_dim(list(trajs.shape[:-2]))
         self.set_times(times)
-        self.set_boundary_conditions(bc_time, bc_pos, bc_vel)
+        self.set_initial_conditions(init_time, init_pos, init_vel)
 
         self.compute_intermediate_terms_single()
         self.compute_intermediate_terms_multi_dof()
@@ -557,7 +607,10 @@ class ProDMP(ProMP):
         #               [*add_dim, num_dof * num_times, num_dof * num_basis_g]
         #            -> [*add_dim, num_dof * num_basis_g, num_dof * num_basis_g]
         A = torch.einsum('...ki,...kj->...ij', pos_H_multi, pos_H_multi)
-        A += torch.eye(self._num_local_params) * reg
+        # todo, check here
+        A += torch.eye(self.num_dof * self.num_basis_g,
+                       dtype=self.dtype,
+                       device=self.device) * reg
 
         # Swap axis and reshape: [*add_dim, num_times, num_dof]
         #                     -> [*add_dim, num_dof, num_times]
@@ -567,8 +620,8 @@ class ProDMP(ProMP):
         #      -> [*add_dim, num_dof * num_times]
         trajs = trajs.reshape([*self.add_dim, -1])
 
-        # Position minus boundary condition terms,
-        pos_wg = trajs - self.pos_bc
+        # Position minus initial condition terms,
+        pos_wg = trajs - self.pos_init
 
         # Einsum_shape: [*add_dim, num_dof * num_times, num_dof * num_basis_g]
         #               [*add_dim, num_dof * num_times]
@@ -587,91 +640,93 @@ class ProDMP(ProMP):
         self.set_mp_params_variances(None)
 
         return {"params": params,
-                "bc_time": bc_time,
-                "bc_pos": bc_pos,
-                "bc_vel": bc_vel}
+                "init_time": init_time,
+                "init_pos": init_pos,
+                "init_vel": init_vel}
 
     def compute_intermediate_terms_single(self):
-        # Determinant of boundary condition,
+        # Determinant of initial condition,
         # Shape: [*add_dim]
-        det = self.y1_bc * self.dy2_bc - self.y2_bc * self.dy1_bc
+        det = self.y1_init * self.dy2_init - self.y2_init * self.dy1_init
         # Compute coefficients to form up traj position and velocity
         # Shape: [*add_dim], [*add_dim, num_times] -> [*add_dim, num_times]
-        xi_1 = torch.einsum("...,...i->...i", self.dy2_bc / det, self.y1) \
-               - torch.einsum("...,...i->...i", self.dy1_bc / det, self.y2)
-        xi_2 = torch.einsum("...,...i->...i", self.y1_bc / det, self.y2) \
-               - torch.einsum("...,...i->...i", self.y2_bc / det, self.y1)
-        xi_3 = torch.einsum("...,...i->...i", self.dy1_bc / det, self.y2) \
-               - torch.einsum("...,...i->...i", self.dy2_bc / det, self.y1)
-        xi_4 = torch.einsum("...,...i->...i", self.y2_bc / det, self.y1) \
-               - torch.einsum("...,...i->...i", self.y1_bc / det, self.y2)
-        dxi_1 = torch.einsum("...,...i->...i", self.dy2_bc / det, self.dy1) \
-                - torch.einsum("...,...i->...i", self.dy1_bc / det, self.dy2)
-        dxi_2 = torch.einsum("...,...i->...i", self.y1_bc / det, self.dy2) \
-                - torch.einsum("...,...i->...i", self.y2_bc / det, self.dy1)
-        dxi_3 = torch.einsum("...,...i->...i", self.dy1_bc / det, self.dy2) \
-                - torch.einsum("...,...i->...i", self.dy2_bc / det, self.dy1)
-        dxi_4 = torch.einsum("...,...i->...i", self.y2_bc / det, self.dy1) \
-                - torch.einsum("...,...i->...i", self.y1_bc / det, self.dy2)
+        xi_1 = torch.einsum("...,...i->...i", self.dy2_init / det, self.y1) \
+               - torch.einsum("...,...i->...i", self.dy1_init / det, self.y2)
+        xi_2 = torch.einsum("...,...i->...i", self.y1_init / det, self.y2) \
+               - torch.einsum("...,...i->...i", self.y2_init / det, self.y1)
+        xi_3 = torch.einsum("...,...i->...i", self.dy1_init / det, self.y2) \
+               - torch.einsum("...,...i->...i", self.dy2_init / det, self.y1)
+        xi_4 = torch.einsum("...,...i->...i", self.y2_init / det, self.y1) \
+               - torch.einsum("...,...i->...i", self.y1_init / det, self.y2)
+        dxi_1 = torch.einsum("...,...i->...i", self.dy2_init / det, self.dy1) \
+                - torch.einsum("...,...i->...i", self.dy1_init / det, self.dy2)
+        dxi_2 = torch.einsum("...,...i->...i", self.y1_init / det, self.dy2) \
+                - torch.einsum("...,...i->...i", self.y2_init / det, self.dy1)
+        dxi_3 = torch.einsum("...,...i->...i", self.dy1_init / det, self.dy2) \
+                - torch.einsum("...,...i->...i", self.dy2_init / det, self.dy1)
+        dxi_4 = torch.einsum("...,...i->...i", self.y2_init / det, self.dy1) \
+                - torch.einsum("...,...i->...i", self.y1_init / det, self.dy2)
 
-        # Generate basis boundary condition values
+        # Generate basis initial condition values
         # [*add_dim, num_basis_g]
-        pos_basis_bc = self.basis_gn.basis(self.bc_time[..., None]).squeeze(-2)
-        vel_basis_bc = self.basis_gn.vel_basis(self.bc_time[..., None]).squeeze(
+        pos_basis_init = self.basis_gn.basis(self.init_time[..., None]).squeeze(
+            -2)
+        vel_basis_init = self.basis_gn.vel_basis(
+            self.init_time[..., None]).squeeze(
             -2)
 
-        # Scale bc_vel
-        bc_vel = self.bc_vel * self.phase_gn.tau[..., None]
+        # Scale init_vel
+        init_vel = self.init_vel * self.phase_gn.tau[..., None]
 
-        # Compute position and velocity boundary condition part
+        # Compute position and velocity initial condition part
         # Einsum shape: [*add_dim, num_times],
         #               [*add_dim, num_dof]
         #            -> [*add_dim, num_dof, num_times]
-        pos_det = torch.einsum('...j,...i->...ij', xi_1, self.bc_pos) \
-                  + torch.einsum('...j,...i->...ij', xi_2, bc_vel)
-        vel_det = torch.einsum('...j,...i->...ij', dxi_1, self.bc_pos) \
-                  + torch.einsum('...j,...i->...ij', dxi_2, bc_vel)
+        pos_det = torch.einsum('...j,...i->...ij', xi_1, self.init_pos) \
+                  + torch.einsum('...j,...i->...ij', xi_2, init_vel)
+        vel_det = torch.einsum('...j,...i->...ij', dxi_1, self.init_pos) \
+                  + torch.einsum('...j,...i->...ij', dxi_2, init_vel)
 
         # Reshape: [*add_dim, num_dof, num_times]
         #       -> [*add_dim, num_dof * num_times]
-        self.pos_bc = torch.reshape(pos_det, [*self.add_dim, -1])
-        self.vel_bc = torch.reshape(vel_det, [*self.add_dim, -1])
+        self.pos_init = torch.reshape(pos_det, [*self.add_dim, -1])
+        self.vel_init = torch.reshape(vel_det, [*self.add_dim, -1])
 
         # Compute position and velocity linear basis part
         # Einsum shape: [*add_dim, num_times],
         #               [*add_dim, num_basis_g]
         #            -> [*add_dim, num_times, num_basis_g]
         self.pos_H_single = \
-            torch.einsum('...i,...j->...ij', xi_3, pos_basis_bc) \
-            + torch.einsum('...i,...j->...ij', xi_4, vel_basis_bc) \
+            torch.einsum('...i,...j->...ij', xi_3, pos_basis_init) \
+            + torch.einsum('...i,...j->...ij', xi_4, vel_basis_init) \
             + self.basis_gn.basis(self.times)
         self.vel_H_single = \
-            torch.einsum('...i,...j->...ij', dxi_3, pos_basis_bc) \
-            + torch.einsum('...i,...j->...ij', dxi_4, vel_basis_bc) \
+            torch.einsum('...i,...j->...ij', dxi_3, pos_basis_init) \
+            + torch.einsum('...i,...j->...ij', dxi_4, vel_basis_init) \
             + self.basis_gn.vel_basis(self.times)
 
     def compute_intermediate_terms_multi_dof(self):
-        # Determinant of boundary condition,
+        # Determinant of initial condition,
         # Shape: [*add_dim]
-        det = self.y1_bc * self.dy2_bc - self.y2_bc * self.dy1_bc
+        det = self.y1_init * self.dy2_init - self.y2_init * self.dy1_init
 
         # Compute coefficients to form up traj position and velocity
         # Shape: [*add_dim], [*add_dim, num_times] -> [*add_dim, num_times]
-        xi_3 = torch.einsum("...,...i->...i", self.dy1_bc / det, self.y2) \
-               - torch.einsum("...,...i->...i", self.dy2_bc / det, self.y1)
-        xi_4 = torch.einsum("...,...i->...i", self.y2_bc / det, self.y1) \
-               - torch.einsum("...,...i->...i", self.y1_bc / det, self.y2)
-        dxi_3 = torch.einsum("...,...i->...i", self.dy1_bc / det, self.dy2) \
-                - torch.einsum("...,...i->...i", self.dy2_bc / det, self.dy1)
-        dxi_4 = torch.einsum("...,...i->...i", self.y2_bc / det, self.dy1) \
-                - torch.einsum("...,...i->...i", self.y1_bc / det, self.dy2)
+        xi_3 = torch.einsum("...,...i->...i", self.dy1_init / det, self.y2) \
+               - torch.einsum("...,...i->...i", self.dy2_init / det, self.y1)
+        xi_4 = torch.einsum("...,...i->...i", self.y2_init / det, self.y1) \
+               - torch.einsum("...,...i->...i", self.y1_init / det, self.y2)
+        dxi_3 = torch.einsum("...,...i->...i", self.dy1_init / det, self.dy2) \
+                - torch.einsum("...,...i->...i", self.dy2_init / det, self.dy1)
+        dxi_4 = torch.einsum("...,...i->...i", self.y2_init / det, self.dy1) \
+                - torch.einsum("...,...i->...i", self.y1_init / det, self.dy2)
 
-        # Generate blocked basis boundary condition values
+        # Generate blocked basis initial condition values
         # [*add_dim, num_dof, num_dof * num_basis_g]
-        pos_basis_bc_multi_dofs = self.basis_gn.basis_multi_dofs(
-            self.bc_time[..., None], self.num_dof)
-        vel_basis_bc_multi_dofs = self.basis_gn.vel_basis_multi_dofs(
-            self.bc_time[..., None], self.num_dof)
+        pos_basis_init_multi_dofs = self.basis_gn.basis_multi_dofs(
+            self.init_time[..., None], self.num_dof)
+        vel_basis_init_multi_dofs = self.basis_gn.vel_basis_multi_dofs(
+            self.init_time[..., None], self.num_dof)
 
         # Compute position and velocity variant part (part 3)
         # Position and velocity part 3_1 and 3_2
@@ -679,22 +734,74 @@ class ProDMP(ProMP):
         #               [*add_dim, num_dof, num_dof * num_basis_g]
         #            -> [*add_dim, num_dof, num_times, num_dof * num_basis_g]
         pos_H_ = torch.einsum('...j,...ik->...ijk', xi_3,
-                              pos_basis_bc_multi_dofs) + \
+                              pos_basis_init_multi_dofs) + \
                  torch.einsum('...j,...ik->...ijk', xi_4,
-                              vel_basis_bc_multi_dofs)
+                              vel_basis_init_multi_dofs)
         vel_H_ = torch.einsum('...j,...ik->...ijk', dxi_3,
-                              pos_basis_bc_multi_dofs) + \
+                              pos_basis_init_multi_dofs) + \
                  torch.einsum('...j,...ik->...ijk', dxi_4,
-                              vel_basis_bc_multi_dofs)
+                              vel_basis_init_multi_dofs)
         # Reshape: [*add_dim, num_dof, num_times, num_dof * num_basis_g]
         #       -> [*add_dim, num_dof * num_times, num_dof * num_basis_g]
-        pos_H_ = \
-            torch.reshape(pos_H_, [*self.add_dim, -1, self._num_local_params])
-        vel_H_ = \
-            torch.reshape(vel_H_, [*self.add_dim, -1, self._num_local_params])
+        # todo, check here
+        pos_H_ = torch.reshape(pos_H_, [*self.add_dim, -1,
+                                        self.num_dof * self.num_basis_g])
+        vel_H_ = torch.reshape(vel_H_, [*self.add_dim, -1,
+                                        self.num_dof * self.num_basis_g])
 
         self.pos_H_multi = \
             pos_H_ + self.basis_gn.basis_multi_dofs(self.times, self.num_dof)
         self.vel_H_multi = \
             vel_H_ + self.basis_gn.vel_basis_multi_dofs(self.times,
                                                         self.num_dof)
+
+    def _show_scaled_basis(self, plot=False) \
+            -> Tuple[torch.Tensor, torch.Tensor]:
+        tau = self.phase_gn.tau
+        delay = self.phase_gn.delay
+        assert tau.ndim == 0 and delay.ndim == 0
+        times = torch.linspace(delay - tau, delay + 2 * tau, steps=1000,
+                               device=self.device, dtype=self.dtype)
+        self.set_add_dim([])
+        self.set_times(times)
+        self.set_initial_conditions(
+            init_time=torch.zeros([], device=self.device,
+                                  dtype=self.dtype) + delay,
+            init_pos=torch.zeros([self.num_dof], device=self.device,
+                                 dtype=self.dtype),
+            init_vel=torch.zeros([self.num_dof], device=self.device,
+                                 dtype=self.dtype),
+        )
+
+        self.compute_intermediate_terms_single()
+
+        weights_goal_scale = self.weights_goal_scale
+
+        # Get basis
+        # Shape: [*add_dim, num_times, num_basis]
+        basis_values = self.pos_H_single * weights_goal_scale
+
+        # Enforce all variables to numpy
+        times, basis_values, delay, tau = \
+            mp_pytorch.util.to_nps(times, basis_values, delay, tau)
+
+        if plot:
+            import matplotlib.pyplot as plt
+            fig, axes = plt.subplots(1, 2, sharex=True, squeeze=False)
+            for i in range(basis_values.shape[-1] - 1):
+                axes[0, 0].plot(times, basis_values[:, i], label=f"w_basis_{i}")
+            axes[0, 0].grid()
+            axes[0, 0].legend()
+            axes[0, 0].axvline(x=delay, linestyle='--', color='k', alpha=0.3)
+            axes[0, 0].axvline(x=delay + tau, linestyle='--', color='k',
+                               alpha=0.3)
+
+            axes[0, 1].plot(times, basis_values[:, -1], label=f"goal_basis")
+            axes[0, 1].grid()
+            axes[0, 1].legend()
+            axes[0, 1].axvline(x=delay, linestyle='--', color='k', alpha=0.3)
+            axes[0, 1].axvline(x=delay + tau, linestyle='--', color='k',
+                               alpha=0.3)
+
+            plt.show()
+        return times, basis_values
